"""A module that parses SigmaStudio project files and extracts adjustable cells."""
import logging
from dataclasses import dataclass
from itertools import dropwhile, takewhile
from typing import ClassVar, List, Union


@dataclass(frozen=True)
class Cell:
    """A cell object is a unit, which represents a cell from SigmaStudio."""

    # The full name of the cell (e.g. 'adjustable_volume_main_left').
    full_name: str

    # The register address of the parameter.
    parameter_address: int

    # The name of the parameter, as defined in SigmaStudio.
    parameter_name: str

    # The value of the parameter, if applicable.
    parameter_value: Union[int, float, None] = None

    # This string separates prefixes within a cell name.
    PREFIX_SEPARATOR: ClassVar[str] = "_"

    # The prefix that marks a cell as adjustable.
    ADJUSTABLE_PREFIX: ClassVar[str] = "adjustable"

    # If a parameter carries this description in its name, it is a register that can be adjusted externally.
    # This is in contrast to parameters that are marked with "alpha".
    TARGET_PARAMETER: ClassVar[str] = "target"

    # The cell name prefix for cells that adjust volume.
    VOLUME_PREFIX: ClassVar[str] = "volume"

    # The cell name for a safety hash.
    SAFETY_HASH: ClassVar[str] = "safety_hash"

    # A complete list of valid prefixes, which are understood by the parser.
    VALID_PREFIX_TOKENS: ClassVar[List[str]] = [ADJUSTABLE_PREFIX, VOLUME_PREFIX]

    @property
    def full_name_tokens(self) -> List[str]:
        """Extract the separated full name, where the separator string is removed.

        For example 'adjustable_volume_main_left' is converted to ['adjustable', 'volume', 'main', 'left']

        Returns:
            List[str]: The list of tokens in the full name.
        """
        return self.full_name.split(Cell.PREFIX_SEPARATOR)

    @property
    def name_tokens(self) -> List[str]:
        """Return the name tokens of the cell without prefix tokens.

        For example ['main', 'left'] for a full cell name 'adjustable_volume_main_left'.

        Returns:
            str: The unprefixed cell name, if prefix tokens exist. The 'full_name_tokens' otherwise.
        """
        if self.prefix_tokens is None:
            return self.full_name_tokens

        return self.full_name_tokens[len(self.prefix_tokens) :]

    @property
    def prefix_tokens(self) -> List[str]:
        """Return only the prefix tokens of the cell.

        For example ['adjustable', 'volume'] for a full cell name 'adjustable_volume_main_left'.

        Returns:
            Union[List[str], None]: The list of tokens, if they are valid prefixes, None otherwise.
        """
        if Cell.PREFIX_SEPARATOR in self.full_name:
            return [prefix for prefix in self.full_name_tokens if prefix in Cell.VALID_PREFIX_TOKENS]

        else:
            return []

    @property
    def is_adjustable(self) -> bool:
        """Determine, whether the cell is adjustable.

        Returns:
            bool: True, if adjustable, False otherwise.
        """
        return Cell.ADJUSTABLE_PREFIX in self.prefix_tokens

    @property
    def is_safety_hash(self) -> bool:
        """Determine, whether the cell contains a safety hash.

        Returns:
            bool: True, if it does, False otherwise.
        """
        return self.full_name == Cell.SAFETY_HASH

    @property
    def is_volume_cell(self) -> bool:
        """Determine, whether the cell is an adjustable volume cell.

        Returns:
            bool: True, if it is an adjustable volume cell, False otherwise.
        """
        return (
            self.is_adjustable
            and (Cell.VOLUME_PREFIX in self.prefix_tokens)
            and (Cell.TARGET_PARAMETER in self.parameter_name)
        )


class Parser:
    """Parse a parameter input file from Sigma Studio and detects cells in it."""

    def __init__(self):
        """Initialize the parameter file parser with an empty list of cells."""
        self.cells: List[Cell] = []

    def extract_cell(self, cell_lines: List[str]) -> Union[Cell, None]:
        """Read a block of data from the parameter listing, and creates a new Cell from it.

        Args:
            cell_lines (List[str]): The lines from the listing to read from.

        Returns:
            Union[Cell, None]: The new Cell, if the data from the listing is valid, None otherwise.
        """
        for line in cell_lines:
            # Assemble a cell from the information in the cell_lines
            split_line = line.split()

            if split_line:
                if split_line[0] == "Cell" and split_line[1] == "Name":

                    name = " ".join(split_line[3:])

                elif split_line[0] == "Parameter":

                    if split_line[1] == "Name":
                        parameter_name = " ".join(split_line[3:])

                    if split_line[1] == "Address":
                        parameter_address = int(split_line[3])

                    if split_line[1] == "Value":
                        data = split_line[3]

                        parameter_value: Union[float, int, None] = None

                        try:
                            parameter_value = int(data)

                        except ValueError:
                            parameter_value = float(data)

        try:
            # Create a new cell, based on the collected parameters
            return Cell(name, parameter_address, parameter_name, parameter_value)

        except NameError:
            # This fails, if any of the required parameters were not extracted successfully.
            return None

    def run(self, file_path: str) -> None:
        """Parse an input file that was exported from Sigma Studio.

        Args:
            file_path (str): The path to the input file
        """
        self.cells.clear()

        if not file_path.endswith(".params"):
            logging.error("The parameter file is not a *.params file! Aborting.")
            return

        # Proceed with opening the file
        try:
            with open(file_path, "r", encoding="utf8") as file:
                logging.info("Using parameter file path %s.", file_path)

                line_iterator = iter(file.readlines())

                while True:
                    # The cascaded iterator looks for blocks within the parameter file, which define a cell.
                    # These blocks are surrounded by empty lines that only contain '\n'.
                    cell_lines = list(takewhile(lambda x: x != "\n", dropwhile(lambda x: x == "\n", line_iterator)))

                    if not cell_lines:
                        # No more blocks to be found in the parameter file.
                        break

                    cell = self.extract_cell(cell_lines)

                    if cell is None:
                        continue

                    elif cell not in self.cells:
                        self.cells.append(cell)

                logging.info("Found a total number of %d unique parameter cells.", len(self.cells))

        except FileNotFoundError:
            logging.info("Parameter file %s not found.", file_path)

    @property
    def safety_hash_cell(self) -> Union[Cell, None]:
        """Find and return the safety hash cell, if it exists.

        Returns:
            Cell: The safety hash cell.
        """
        safety_hash_cells = [cell for cell in self.cells if cell.is_safety_hash]

        if 1 != len(safety_hash_cells):
            # There must be exactly one safety hash cell.
            return None

        else:
            return safety_hash_cells[0]

    @property
    def volume_cells(self) -> List[Cell]:
        """Return all cells that can be used for volume adjustment. These are user defined with a certain name pattern.

        Returns:
            List[Cell]: The list of adjustable volume cells
        """
        return [cell for cell in self.cells if cell.is_volume_cell]

    def get_matching_cells_by_name_tokens(self, all_cells: List[Cell], name_tokens: List[str]) -> List[Cell]:
        """Find cells in a list of cells, whose names match the specified name tokens.

        Args:
            all_cells (List[Cell]): The list of cells to parse.
            name_tokens (List[str]): The tokens to check against.

        Returns:
            List[Cell]: The matched cells
        """
        return [cell for cell in all_cells if name_tokens == cell.name_tokens]



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import control_pb2 as control__pb2


class BackendStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.control = channel.unary_unary(
                '/sigmadsp.backend_service.Backend/control',
                request_serializer=control__pb2.ControlRequest.SerializeToString,
                response_deserializer=control__pb2.ControlResponse.FromString,
                )
        self.control_parameter = channel.unary_unary(
                '/sigmadsp.backend_service.Backend/control_parameter',
                request_serializer=control__pb2.ControlParameterRequest.SerializeToString,
                response_deserializer=control__pb2.ControlResponse.FromString,
                )


class BackendServicer(object):
    """Missing associated documentation comment in .proto file."""

    def control(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def control_parameter(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_BackendServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'control': grpc.unary_unary_rpc_method_handler(
                    servicer.control,
                    request_deserializer=control__pb2.ControlRequest.FromString,
                    response_serializer=control__pb2.ControlResponse.SerializeToString,
            ),
            'control_parameter': grpc.unary_unary_rpc_method_handler(
                    servicer.control_parameter,
                    request_deserializer=control__pb2.ControlParameterRequest.FromString,
                    response_serializer=control__pb2.ControlResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'sigmadsp.backend_service.Backend', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Backend(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def control(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/sigmadsp.backend_service.Backend/control',
            control__pb2.ControlRequest.SerializeToString,
            control__pb2.ControlResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def control_parameter(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/sigmadsp.backend_service.Backend/control_parameter',
            control__pb2.ControlParameterRequest.SerializeToString,
            control__pb2.ControlResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



"""This module implements an SPI handler that talks to Sigma DSP devices."""
import logging
import threading
from multiprocessing import Pipe

import spidev


def build_spi_frame(address: int, data: bytes) -> bytearray:
    """Build an SPI frame that is later written to the DSP.

    Args:
        address (int): The register address that the data is written to
        data (bytes): The register content

    Returns:
        bytearray: The complete SPI frame buffer
    """
    frame = bytearray(SpiHandler.HEADER_LENGTH)
    frame[0] = SpiHandler.WRITE
    frame[1:3] = address.to_bytes(SpiHandler.ADDRESS_LENGTH, "big")
    frame += data

    return frame


class SpiHandler:
    """Handle SPI transfers from and to SigmaDSP chipsets.

    Tested with ADAU145X
    """

    # Length of addresses (in bytes) for accessing registers
    ADDRESS_LENGTH = 2

    # Length of the SPI header for communicating with SigmaDSP chipsets
    HEADER_LENGTH = 3

    # Maximum number of bytes per SPI transfer
    MAX_SPI_BYTES = 4096

    # Maximum number of words (32 bit) that can be transferred with the
    # maximum number of allowed bytes per SPI transfer
    MAX_PAYLOAD_WORDS = int((MAX_SPI_BYTES - HEADER_LENGTH) / 4)

    # Derive maximum payload (bytes) from number of maximum words
    MAX_PAYLOAD_BYTES = MAX_PAYLOAD_WORDS * 4

    WRITE = 0
    READ = 1

    def __init__(self):
        """Initialize the SpiHandler thread."""
        self._initialize_spi()

        # Generate a Pipe, for communicating with the SPI handler thread within this class.
        self.pipe_end_owner, self.pipe_end_user = Pipe()

        logging.info("Starting SPI handling thread.")
        self.thread = threading.Thread(target=self.serve_forever, name="SPI handler thread")
        self.thread.daemon = True
        self.thread.start()

    def _initialize_spi(self, bus: int = 0, device: int = 0):
        """Initialize the SPI hardware.

        Bus and device numbers shall be kept at (0, 0) for RaspberryPi hardware.
        The RaspberryPi only has one SPI peripheral.

        Args:
            bus (int, optional): Bus number. Defaults to 0.
            device (int, optional): Device number. Defaults to 0.
        """
        self.spi = spidev.SpiDev()
        self.spi.open(bus, device)

        # The SigmaDSP allows a maximum SPI transfer speed of 20 MHz.
        # Raspberry Pi hardware allows binary steps (1 MHz, 2 MHz, ...)
        self.spi.max_speed_hz = 16000000

        # The SigmaDSP uses SPI mode 0 with 8 bits per word. Do not change.
        self.spi.mode = 0
        self.spi.bits_per_word = 8

    def write(self, address: int, data: bytes):
        """Write data over the SPI interface via the pipe to the SPI thread.

        Args:
            address (int): DSP register address to write to
            data (bytes): Binary data to write
        """
        self.pipe_end_owner.send("write")
        self.pipe_end_owner.send((address, data))

    def read(self, address: int, length: int) -> bytes:
        """Read data from the SPI interface via the pipe to the SPI thread.

        Args:
            address (int): DSP register address to read from
            length (int): Number of bytes to read

        Returns:
            bytes: Register content
        """
        self.pipe_end_owner.send("read")
        self.pipe_end_owner.send((address, length))

        data = self.pipe_end_owner.recv()

        return data

    def serve_forever(self):
        """Handle incoming requests for writing or reading data over SPI."""
        while True:
            mode = self.pipe_end_user.recv()

            if mode == "write":
                address, data = self.pipe_end_user.recv()
                self._write_spi(address, data)

            elif mode == "read":
                address, length = self.pipe_end_user.recv()
                data = self._read_spi(address, length)
                self.pipe_end_user.send(data)

    def _read_spi(self, address: int, length: int) -> bytes:
        """Read data over the SPI port from a SigmaDSP.

        Args:
            address (int): Address to read from
            length (int): Number of bytes to read

        Returns:
            bytes: Data that was read from the DSP
        """
        spi_request = bytearray(length + SpiHandler.HEADER_LENGTH)
        spi_request[0] = SpiHandler.READ
        spi_request[1:3] = address.to_bytes(SpiHandler.ADDRESS_LENGTH, "big")

        # Read, by writing zeros
        spi_response = self.spi.xfer(spi_request)

        return bytes(spi_response[SpiHandler.HEADER_LENGTH :])

    def _write_spi(self, address: int, data: bytes):
        """Write data over the SPI port onto a SigmaDSP.

        Args:
            address (int): Address to write to
            data (bytes): Data to write

        Returns:
            int: Number of bytes written
        """
        remaining_data_length = len(data)
        current_address = address
        current_data = data

        while remaining_data_length > 0:
            # There is data remaining for writing

            if remaining_data_length >= SpiHandler.MAX_PAYLOAD_BYTES:
                # Packet has to be split into smaller chunks,
                # where the write address is advanced accordingly.
                # DSP register addresses are counted in words (32 bit per increment).

                # Build the frame from a subset of the input data, and write it
                frame = build_spi_frame(
                    current_address,
                    current_data[: SpiHandler.MAX_PAYLOAD_BYTES],
                )
                self.spi.writebytes(frame)

                # Update address, data counter, and the binary data buffer
                current_address += SpiHandler.MAX_PAYLOAD_WORDS
                remaining_data_length -= SpiHandler.MAX_PAYLOAD_BYTES
                current_data = current_data[SpiHandler.MAX_PAYLOAD_BYTES :]

            else:
                # The packet fits into one transmission.
                frame = build_spi_frame(current_address, current_data)
                self.spi.writebytes(frame)
                remaining_data_length = 0



"""This module is the frontend to the SigmaDSP backend service.

It can control the backend via the command line.
"""
import argparse
import logging

import grpc

from sigmadsp.generated.backend_service.control_pb2 import (
    ControlParameterRequest,
    ControlRequest,
    ControlResponse,
)
from sigmadsp.generated.backend_service.control_pb2_grpc import BackendStub


def main():
    """The main frontend command-line application, which controls the SigmaDSP backend."""
    logging.basicConfig(level=logging.INFO)

    argument_parser = argparse.ArgumentParser()

    argument_parser.add_argument(
        "-p",
        "--port",
        required=False,
        type=int,
        help="Set the port, on which the backend listens for requests.",
    )

    argument_parser.add_argument(
        "-a",
        "--address",
        required=False,
        type=str,
        help="Set the IP address, on which the backend listens for requests.",
    )

    argument_parser.add_argument(
        "-av",
        "--adjust_volume",
        required=False,
        type=float,
        help="Adjust the volume by a certain value in dB (positive or negative).",
    )

    argument_parser.add_argument(
        "-sv",
        "--set_volume",
        required=False,
        type=float,
        help="Sets the volume to a certain value in dB (zero or lower).",
    )

    argument_parser.add_argument(
        "-r",
        "--reset",
        required=False,
        help="Soft-reset the DSP.",
        action="store_true",
    )

    argument_parser.add_argument(
        "-R",
        "--hard-reset",
        required=False,
        help="hard-reset the DSP.",
        action="store_true",
    )

    argument_parser.add_argument(
        "-lp",
        "--load_parameters",
        required=False,
        help="Load new parameter file",
    )

    arguments = argument_parser.parse_args()

    backend_port = 50051
    backend_address = "localhost"

    if arguments.port is not None:
        backend_port = arguments.port

    if arguments.address is not None:
        backend_address = arguments.address

    response: ControlResponse
    control_request = ControlRequest()
    control_parameter_request = ControlParameterRequest()

    with grpc.insecure_channel(f"{backend_address}:{backend_port}") as channel:
        stub = BackendStub(channel)

        if arguments.adjust_volume is not None:
            control_parameter_request.change_volume.name_tokens[:] = ["main"]
            control_parameter_request.change_volume.value = arguments.adjust_volume
            control_parameter_request.change_volume.relative = True

            response = stub.control_parameter(control_parameter_request)

        if arguments.set_volume is not None:
            control_parameter_request.change_volume.name_tokens[:] = ["main"]
            control_parameter_request.change_volume.value = arguments.set_volume
            control_parameter_request.change_volume.relative = False

            response = stub.control_parameter(control_parameter_request)

        if arguments.load_parameters is not None:
            with open(arguments.load_parameters, "r", encoding="utf8") as parameter_file:
                control_request.load_parameters.content[:] = parameter_file.readlines()

            response = stub.control(control_request)

        if arguments.reset is True:
            control_request.reset_dsp = True

            response = stub.control(control_request)

        if arguments.hard_reset is True:
            control_request.hard_reset_dsp = True

            response = stub.control(control_request)

    logging.info(response.message)


if __name__ == "__main__":
    main()



"""This module communicates with SigmaStudio.

It can receive read/write requests and return with read response packets.
"""
import logging
import socket
import socketserver
import threading
from dataclasses import dataclass
from multiprocessing import Pipe

from sigmadsp.helper.conversion import (
    bytes_to_int8,
    bytes_to_int16,
    bytes_to_int32,
    int8_to_bytes,
    int16_to_bytes,
    int32_to_bytes,
)


@dataclass(frozen=True)
class WriteRequest:
    """SigmaStudio requests to write data to the DSP."""

    address: int
    data: bytes


@dataclass(frozen=True)
class ReadRequest:
    """SigmaStudio requests to read data from the DSP."""

    address: int
    length: int


@dataclass(frozen=True)
class ReadResponse:
    """SigmaStudio is sent a response to its ReadRequest."""

    data: bytes


class ThreadedTCPServer(socketserver.ThreadingTCPServer):
    """The threaded TCP server that is used for communicating with SigmaStudio.

    It will be instantiated by SigmaStudioInterface. General server settings can be adjusted below.
    """

    allow_reuse_address = True

    def __init__(self, *args, **kwargs):
        """Initialize the ThreadedTCPServer with a Pipe for communicating with the TCP server worker thread."""
        self.pipe_end_owner, self.pipe_end_user = Pipe()

        super().__init__(*args, **kwargs)


class SigmaStudioRequestHandler(socketserver.BaseRequestHandler):
    """Request handler for messages from SigmaStudio."""

    request: socket.socket
    server: ThreadedTCPServer

    HEADER_LENGTH: int = 14
    COMMAND_WRITE: int = 0x09
    COMMAND_READ: int = 0x0A
    COMMAND_READ_RESPONSE: int = 0x0B

    def receive_amount(self, total_size: int) -> bytearray:
        """Receives data from a request, until the desired size is reached.

        Args:
            total_size (int): The payload size in bytes to get.

        Returns:
            bytearray: The received data.
        """
        payload_data = bytearray()

        while total_size > len(payload_data):
            # Wait until the complete TCP payload was received.
            received = self.request.recv(total_size - len(payload_data))

            if 0 == len(received):
                # Give up, if no more data arrives.
                # Close the socket.
                self.request.shutdown(socket.SHUT_RDWR)
                self.request.close()

                raise ConnectionError

            payload_data.extend(received)

        return payload_data

    def handle_write_data(self, packet_header: bytes):
        """Handle requests, where SigmaStudio wants to write to the DSP.

        Args:
            packet_header (bytes): The binary request header, as received from SigmaStudio.
        """
        # This field indicates whether the packet is a block write or a safeload write.
        # TODO: This field is currently unused.
        block_safeload = bytes_to_int8(packet_header, 1)
        del block_safeload

        # This indicates the channel number.
        # TODO: This field is currently unused.
        channel_number = bytes_to_int8(packet_header, 2)
        del channel_number

        # This indicates the total length of the write packet (uint32).
        # This field is currently unused.
        total_length = bytes_to_int32(packet_header, 3)
        del total_length

        # The address of the chip to which the data has to be written.
        # The chip address is appended by the read/write bit: shifting right by one bit removes it.
        # TODO: This field is currently unused.
        chip_address = bytes_to_int8(packet_header, 7) >> 1
        del chip_address

        # The length of the data (uint32).
        total_payload_size = bytes_to_int32(packet_header, 8)

        # The address of the module whose data is being written to the DSP (uint16).
        address = bytes_to_int16(packet_header, 12)

        payload_data = self.receive_amount(total_payload_size)

        self.server.pipe_end_owner.send(WriteRequest(address, payload_data))

    def handle_read_request(self, packet_header: bytes):
        """Handle requests, where SigmaStudio wants to read from the DSP.

        Args:
            packet_header (bytes): The binary request header, as received from SigmaStudio.
        """
        # This indicates the total length of the read packet (uint32)
        total_length = bytes_to_int32(packet_header, 1)
        del total_length

        # The address of the chip from which the data has to be read
        chip_address = bytes_to_int8(packet_header, 5)

        # The length of the data (uint32)
        data_length = bytes_to_int32(packet_header, 6)

        # The address of the module whose data is being read from the DSP (uint16)
        address = bytes_to_int16(packet_header, 10)

        # Notify application of read request
        self.server.pipe_end_owner.send(ReadRequest(address, data_length))

        # Wait for payload data that goes into the read response
        read_response = self.server.pipe_end_owner.recv()

        payload_data = read_response.data

        # Build the read response packet, starting with length calculations ...
        total_transmit_length = SigmaStudioRequestHandler.HEADER_LENGTH + data_length
        transmit_data = bytearray(total_transmit_length)

        # ... followed by populating the byte fields.
        int8_to_bytes(
            SigmaStudioRequestHandler.COMMAND_READ_RESPONSE,
            transmit_data,
            0,
        )
        int32_to_bytes(total_transmit_length, transmit_data, 1)
        int8_to_bytes(chip_address, transmit_data, 5)
        int32_to_bytes(data_length, transmit_data, 6)
        int16_to_bytes(address, transmit_data, 10)

        # Success (0)/failure (1) byte at pos. 12. Always assume success.
        int16_to_bytes(0, transmit_data, 12)

        # Reserved zero byte at pos. 13
        int16_to_bytes(0, transmit_data, 13)

        transmit_data[SigmaStudioRequestHandler.HEADER_LENGTH :] = payload_data

        self.request.sendall(transmit_data)

    def handle(self):
        """Call, when the TCP server receives new data for handling.

        It never stops, except if the connection is reset.
        """
        while True:
            try:
                # Receive the packet header in the beginning.
                packet_header = self.receive_amount(SigmaStudioRequestHandler.HEADER_LENGTH)

                # The first byte of the header contains the command from SigmaStudio.
                command = packet_header[0]

                if command == SigmaStudioRequestHandler.COMMAND_WRITE:
                    self.handle_write_data(packet_header)

                elif command == SigmaStudioRequestHandler.COMMAND_READ:
                    self.handle_read_request(packet_header)

                else:
                    logging.info("Received an unknown command code '%s'.", hex(command))

            except ConnectionError:
                break


class SigmaStudioInterface:
    """This is an interface class for communicating with SigmaStudio.

    It creates a TCP server, which SigmaStudio talks to.
    """

    def __init__(self, host: str, port: int):
        """Initialize the SigmaStudio interface.

        Starts the main TCP worker and initializes a pipe for communicating with the sigmadsp backend.

        Args:
            host (str): Listening IP address
            port (int): Port to listen at
        """
        self.host = host
        self.port = port

        # Generate a Pipe for communicating with the TCP server worker thread within this class.
        self.pipe_end_owner, self.pipe_end_user = Pipe()

        tcp_server_worker_thread = threading.Thread(target=self.tcp_server_worker, name="TCP server worker thread")
        tcp_server_worker_thread.daemon = True
        tcp_server_worker_thread.start()

    def tcp_server_worker(self):
        """The main worker for the TCP server."""
        tcp_server = ThreadedTCPServer((self.host, self.port), SigmaStudioRequestHandler)

        with tcp_server:
            # Base TCP server thread
            # This initial thread starts one more thread for each request.
            tcp_server_thread = threading.Thread(target=tcp_server.serve_forever, name="TCP server thread")
            tcp_server_thread.daemon = True
            tcp_server_thread.start()

            while True:
                # Wait for a request from the TCP server
                request = tcp_server.pipe_end_user.recv()

                if isinstance(request, WriteRequest):
                    # Write request received, don't do anything else
                    self.pipe_end_owner.send(request)

                elif isinstance(request, ReadRequest):
                    # Read request received, wait for data to send to PC application
                    self.pipe_end_owner.send(request)

                    read_response = self.pipe_end_owner.recv()

                    # Send read response
                    tcp_server.pipe_end_user.send(read_response)



"""The backend application of sigmadsp.

This module includes the main application that provides an interface between the TCP server that faces SigmaStudio,
and an SPI handler that controls a DSP.

Commands from Sigma Studio are received, and translated to SPI read/write requests.
"""
import argparse
import logging
import sys
import threading
from concurrent.futures import ThreadPoolExecutor
from typing import List

import grpc
import yaml

import sigmadsp
from sigmadsp.communication.sigma_tcp_server import (
    ReadRequest,
    ReadResponse,
    SigmaStudioInterface,
    WriteRequest,
)
from sigmadsp.generated.backend_service.control_pb2 import (
    ControlParameterRequest,
    ControlRequest,
    ControlResponse,
)
from sigmadsp.generated.backend_service.control_pb2_grpc import (
    BackendServicer,
    add_BackendServicer_to_server,
)
from sigmadsp.hardware.adau14xx import Adau14xx
from sigmadsp.hardware.spi import SpiHandler
from sigmadsp.helper.parser import Parser


class SigmadspSettings:
    """This class holds and manages settings for the SigmaDSP application."""

    default_config_path = "/var/lib/sigmadsp/config.yaml"

    def __init__(self, config_path: str = None):
        """Load a config file in *.yaml format from a specified path.

        If no file is provided, the default path is used for loading settings.

        Args:
            config_path (str, optional): The input path of the settings file.
                Defaults to None.
        """
        if config_path is None:
            config_path = SigmadspSettings.default_config_path

        try:
            # Open settings file, in order to configure the application
            with open(config_path, "r", encoding="utf8") as settings_file:
                self.config = yaml.safe_load(settings_file)
                logging.info("Settings file %s was loaded.", config_path)

        except FileNotFoundError:
            logging.error("Settings file not found at %s. Aborting.", config_path)
            raise

        self.load_parameters()

    def load_parameters(self) -> None:
        """Load parameter cells, according to the parameter file path that is defined in the settings object."""
        parser = Parser()
        parser.run(self.config["parameters"]["path"])

        self.parameter_parser = parser

    def store_parameters(self, lines: List[str]):
        """Store parameters to the parameter file.

        Args:
            lines (List[str]): [description]
        """
        with open(self.config["parameters"]["path"], "w", encoding="UTF8") as parameter_file:
            parameter_file.writelines(lines)

        self.load_parameters()


class BackendService(BackendServicer):
    """The backend service that handles the underlying TCP server and SPI handler.

    This service also reacts to rpyc remote procedure calls, for performing actions with the DSP over SPI.
    """

    def __init__(self, settings: SigmadspSettings):
        """Initialize service and start all relevant threads (TCP, SPI).

        Args:
            settings (SigmadspSettings): The settings object.
        """
        super().__init__()

        # If configuration_unlocked is False, no DSP parameters can be changed.
        self.configuration_unlocked: bool = False

        self.settings = settings

        # Create an SPI handler, along with its thread
        self.spi_handler = SpiHandler()

        # Create a SigmaTCPServer, along with its various threads
        self.sigma_tcp_server = SigmaStudioInterface(
            self.settings.config["host"]["ip"],
            self.settings.config["host"]["port"],
        )
        logging.info(
            "Sigma TCP server started on [%s]:%d.",
            self.settings.config["host"]["ip"],
            self.settings.config["host"]["port"],
        )

        # Create the worker thread for the handler itself
        worker_thread = threading.Thread(target=self.worker, name="Backend service worker thread")
        worker_thread.daemon = True
        worker_thread.start()

        logging.info("Specified DSP type is '%s'.", self.settings.config["dsp"]["type"])

        if self.settings.config["dsp"]["type"] == "adau14xx":
            self.dsp = Adau14xx(self.settings.config, self.spi_handler)

        else:
            logging.error(
                "DSP type '%s' is not known! Aborting.",
                self.settings.config["dsp"]["type"],
            )
            sys.exit(0)

        self.safety_check()

    def safety_check(self) -> None:
        """Check a hash cell within the DSP's memory against a hash value from the parameter file.

        Only if they match, configuration of DSP parameters is allowed. Other operations (e.g. reset, or
        loading a new parameter file) are still allowed, even in a locked configuration state.
        """
        safety_hash_cell = self.settings.parameter_parser.safety_hash_cell

        if safety_hash_cell is None:
            logging.warning("Safety hash in cell not present! Configuration locked.")
            self.configuration_unlocked = False

        else:
            dsp_hash = self.dsp.get_parameter_value(safety_hash_cell.parameter_address, data_format="int")

            if safety_hash_cell.parameter_value != dsp_hash:
                logging.warning("Safety hash in cell does not match! Configuration locked.")
                self.configuration_unlocked = False

            else:
                logging.info("Safety check successful. Configuration unlocked.")
                self.configuration_unlocked = True

    def worker(self):
        """Main worker functionality.

        Gets requests from the TCP server component and forwards them to the SPI handler.
        """
        while True:
            request = self.sigma_tcp_server.pipe_end_user.recv()

            if isinstance(request, WriteRequest):
                self.spi_handler.write(request.address, request.data)

            elif isinstance(request, ReadRequest):
                payload = self.spi_handler.read(request.address, request.length)

                self.sigma_tcp_server.pipe_end_user.send(ReadResponse(payload))

    def control_parameter(self, request: ControlParameterRequest, context):
        """Main backend entry point for control messages that change or read parameters.

        Here, the configuration has to be in an unlocked state, which requires a hash number in the DSP firmware to
        match the hash number that is provided in the parameter file.

        Args:
            request (ControlParameterRequest): The request that the backend shall handle.
            context (Any): The context within which to handle the request (unused).

        Returns:
            ControlResponse: A control response message, returned to the caller.
        """
        response = ControlResponse()

        # Determine the type of control request.
        command = request.WhichOneof("command")
        response.success = False

        if not self.configuration_unlocked:
            # Do not allow to change parameters.
            response.message = "Configuration locked, parameters cannot be changed."
            return response

        if "change_volume" == command:
            volume_cells_to_adjust = self.settings.parameter_parser.get_matching_cells_by_name_tokens(
                self.settings.parameter_parser.volume_cells, list(request.change_volume.name_tokens)
            )

            if not volume_cells_to_adjust:
                response.message = (
                    f"No volume cell identified by {request.change_volume.name_tokens} was found."
                    "Available controls are "
                    f"{([cell.name_tokens for cell in self.settings.parameter_parser.volume_cells])}."
                )

            for volume_cell in volume_cells_to_adjust:
                if request.change_volume.relative:
                    new_volume_db = self.dsp.adjust_volume(
                        request.change_volume.value,
                        volume_cell.parameter_address,
                    )

                elif not request.change_volume.relative:
                    new_volume_db = self.dsp.set_volume(
                        request.change_volume.value,
                        volume_cell.parameter_address,
                    )

                response.message = f"Set volume of cell '{volume_cell.full_name}' to {new_volume_db:.2f} dB."

        response.success = True
        return response

    def control(self, request: ControlRequest, context):
        """Main backend entry point for control messages.

        Args:
            request (ControlRequest): The request that the backend shall handle.
            context (Any): The context within which to handle the request (unused).

        Returns:
            ControlResponse: A control response message, returned to the caller.
        """
        response = ControlResponse()

        # Determine the type of control request.
        command = request.WhichOneof("command")

        if "reset_dsp" == command:
            self.dsp.soft_reset()
            response.message = "Reset DSP."

        elif "hard_reset_dsp" == command:
            self.dsp.hard_reset()
            response.message = "Hard-reset DSP."

        elif "load_parameters" == command:
            self.settings.store_parameters(list(request.load_parameters.content))

            # Repeat safety check after loading a new set of parameters
            self.safety_check()

            if self.configuration_unlocked:
                response.message = "Loaded parameters, control is unlocked."

            else:
                response.message = "Safety check failed, parameters cannot be adjusted."

        response.success = True
        return response


def launch(settings: SigmadspSettings):
    """Launch the backend application.

    Args:
        settings (SigmadspSettings): Settings object for the backend application.
    """
    # Create the backend service, a grpc service
    grpc_server = grpc.server(ThreadPoolExecutor(max_workers=10))

    backend_port = settings.config["backend"]["port"]
    backend_address = f"[::]:{backend_port}"

    add_BackendServicer_to_server(BackendService(settings), grpc_server)
    grpc_server.add_insecure_port(backend_address)
    grpc_server.start()

    logging.info("Backend service started on %s", backend_address)

    grpc_server.wait_for_termination()


def main():
    """Launch the backend with default settings."""
    logging.basicConfig(level=logging.INFO)

    logging.info("Starting the sigmadsp backend, version %s.", sigmadsp.__version__)

    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument(
        "-s",
        "--settings",
        required=False,
        help=(
            "specifies the settings file path for configuring ",
            "the sigmadsp backend application",
        ),
    )
    arguments = argument_parser.parse_args()

    settings = SigmadspSettings(arguments.settings)
    launch(settings)


if __name__ == "__main__":
    main()



"""General definitions for interfacing DSPs."""
import time
from dataclasses import dataclass
from typing import List, Union

import gpiozero

from sigmadsp.hardware.spi import SpiHandler


@dataclass
class Pin:
    """A class that describes a general DSP pin."""

    name: str
    number: int

    def __post_init__(self):
        """Initialize the generic device, based on the configured parameters."""
        self.control = gpiozero.GPIODevice(self.number)


@dataclass
class InputPin(Pin):
    """A class that describes a DSP input pin."""

    pull_up: bool
    active_state: bool
    bounce_time: Union[float, None]

    def __post_init__(self):
        """Initialize the input device, based on the configured parameters."""
        self.control = gpiozero.DigitalInputDevice(self.number, self.pull_up, self.active_state, self.bounce_time)


@dataclass
class OutputPin(Pin):
    """A class that describes a DSP output pin."""

    initial_value: bool
    active_high: bool

    def __post_init__(self):
        """Initialize the output device, based on the configured parameters."""
        self.control = gpiozero.DigitalOutputDevice(self.number, self.active_high, self.initial_value)


class Dsp:
    """A generic DSP class, to be extended by child classes."""

    def __init__(self, config: dict, spi_handler: SpiHandler):
        """Initialize the DSP with an SpiHandler that talks to it.

        Args:
            config (dict): Configuration settings, from the general configuration file.
            spi_handler (SpiHandler): The SpiHandler that communicates with the DSP.
        """
        self.config = config
        self.pins: List[Pin] = []
        self.parse_config()

        self.spi_handler = spi_handler
        self.hard_reset()

    def parse_config(self):
        """Parse the configuration file and extract relevant information."""
        for pin_definition_key in self.config["dsp"]["pins"]:
            pin_definition = self.config["dsp"]["pins"][pin_definition_key]

            if pin_definition["mode"] == "output":
                output_pin = OutputPin(
                    pin_definition_key,
                    pin_definition["number"],
                    pin_definition["initial_state"],
                    pin_definition["active_high"],
                )

                self.add_pin(output_pin)

            elif pin_definition["mode"] == "input":
                input_in = InputPin(
                    pin_definition_key,
                    pin_definition["number"],
                    pin_definition["pull_up"],
                    pin_definition["active_state"],
                    pin_definition["bounce_time"],
                )

                self.add_pin(input_in)

    def get_pin_by_name(self, name: str) -> Union[Pin, None]:
        """Get a pin by its name.

        Args:
            name (str): The name of the pin.

        Returns:
            Union[Pin, None]: The pin, if one matches, None otherwise.
        """
        for pin in self.pins:
            if pin.name == name:
                return pin

        return None

    def has_pin(self, pin: Pin) -> bool:
        """Check, if a pin is known in the list of pins.

        Args:
            pin (Pin): The pin to look for.

        Returns:
            bool: True, if the pin is known, False otherwise.
        """
        return bool(self.get_pin_by_name(pin.name) is not None)

    def add_pin(self, pin: Pin):
        """Add a pin to the list of pins, if it doesn't exist yet.

        Args:
            pin (Pin): The pin to add.
        """
        if not self.has_pin(pin):
            self.pins.append(pin)

    def remove_pin_by_name(self, name: str):
        """Remove a pin, based on its name.

        Args:
            name (str): The pin name.
        """
        pin = self.get_pin_by_name(name)

        if pin:
            self.pins.remove(pin)

    def hard_reset(self, delay: float = 0):
        """Hard reset the DSP.

        Set and release the corresponding pin for resetting.
        """
        pin = self.get_pin_by_name("reset")

        if not pin:
            return

        pin.control.on()
        time.sleep(delay)
        pin.control.off()



"""Installs the SigmaDSP package."""
import os
import sys

import setuptools

# Add current folder to path
# This is required to import versioneer in an isolated pip build
# Prepending allows not to break on a non-isolated build when versioneer
# is already installed.
# (c.f. https://github.com/scikit-build/cmake-python-distributions/issues/171)
sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))

import versioneer  # noqa: E402 # pylint: disable=wrong-import-position

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    name="sigmadsp",
    author="Adrian Figueroa",
    author_email="elagil@takanome.de",
    description="A package for controlling Analog Devices Sigma DSP chipsets.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/elagil/sigmadsp",
    project_urls={
        "Bug Tracker": "https://github.com/elagil/sigmadsp/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=["pyyaml", "spidev", "grpcio", "protobuf", "RPi.GPIO"],
    entry_points={
        "console_scripts": [
            "sigmadsp-backend=sigmadsp.application.backend:main",
            "sigmadsp=sigmadsp.application.frontend:main",
        ],
    },
)
